(**************************************************************************)
(*                                                                        *)
(*  FORK                                                                  *)
(*                                                                        *)
(*  Author:  Franois Pottier, INRIA Paris-Rocquencourt                   *)
(*  Version: 20101109                                                     *)
(*                                                                        *)
(*  The copyright to this code is held by Institut National de Recherche  *)
(*  en Informatique et en Automatique (INRIA). All rights reserved. This  *)
(*  file is distributed under the license CeCILL-C (see file LICENSE).    *)
(*                                                                        *)
(**************************************************************************)

(* Fragments. *)

(* ------------------------------------------------------------------------- *)

(* A fragment is a type with a hole: typically, a tuple type with a hole at
   its end. This is represented, in a traditional way, as a type operator,
   so that application corresponds to filling the hole. *)

kind fragment = * -> *

(* ------------------------------------------------------------------------- *)

(* Conventions. *)

(* In this prototype implementation, by default, every lambda-bound type
   variable must be annotated with a kind. In order to reduce the burden,
   default kind declarations are permitted. Here, for instance, every
   type variable whose name begins with [tail] (and optionally ends with
   some underscores, primes, and digits) is implicitly declared to have
   kind [*]. *)

default kind f : fragment

default kind tail : *

default kind data : *

(* ------------------------------------------------------------------------- *)

(* The empty fragment is the identity function. *)

type fnil : fragment =
  \tail. tail

(* ------------------------------------------------------------------------- *)

(* The singleton fragment [fcell data] is a product of [data] and the hole. *)

type fcell : * -> fragment =
  \data. \tail. (data, tail)

(* ------------------------------------------------------------------------- *)

(* Fragment concatenation is function composition. *)

type @ : fragment -> fragment -> fragment =
  \f1 f2 tail. f1 (f2 tail)

(* Properties of concatenation. *)

lemma @_left_unit:
  forall f.
  fnil `@` f = f

lemma @_right_unit:
  forall f.
  f `@` fnil = f

lemma @_associative:
  forall f1 f2 f3.
  (f1 `@` f2) `@` f3 = f1 `@` (f2 `@` f3)

(* ------------------------------------------------------------------------- *)

(* Extension is concatenation with a singleton. *)

type snoc : fragment -> * -> fragment =
  \f. \data. f `@` fcell data

lemma snoc_alternative_definition:
  forall f data.
  f `snoc` data = (\tail. f (data, tail))

