(**************************************************************************)
(*                                                                        *)
(*  FORK                                                                  *)
(*                                                                        *)
(*  Author:  Franois Pottier, INRIA Paris-Rocquencourt                   *)
(*  Version: 20101109                                                     *)
(*                                                                        *)
(*  The copyright to this code is held by Institut National de Recherche  *)
(*  en Informatique et en Automatique (INRIA). All rights reserved. This  *)
(*  file is distributed under the license CeCILL-C (see file LICENSE).    *)
(*                                                                        *)
(**************************************************************************)

(* Arrays. *)

(* This is an implementation of arrays in F_omega. The recursive sorts and
   types of Fork are not exploited here. *)

(* This prototype implementation of Fork does not have true modularity. The
   interface of this array implementation consists of the types and terms
   whose names begin with [array] or [index]. These types and terms are to be
   viewed as abstract, even though this is not enforced. The other
   definitions, whose names begin with [__], are to be considered as internal,
   even though this is not enforced. *)

(* ------------------------------------------------------------------------- *)

(* Public types and terms. *)

(* A type constructor for arrays. *)

type array : later fragment -> * =
  (* an array is a pair of the array data and an end address *)
  \f : later fragment. (__fheap f, __faddress f)

(* A type constructor for array indices. *)

type index : later fragment -> later * -> * =
  (* an index is a pair of functions that read/write an array at this index *)
  \f1 : later fragment. \data : later *.
    (
      forall f2. array (f1 `@` f2) -> data,
      forall f2. array (f1 `@` f2) -> data -> array (f1 `@` f2)
    )

(* The empty array. *)

term array_empty : array fnil =
  ((), __zero)

(* Array extension. *)

term array_extend : forall f data. array f -> data -> array (f `snoc` data) =
  /\ f data.
  \s : array f.
  \d : data.
    let (h, addr) = s in
    let (_, set) = addr in
    (
      set [()] [fcell data ()] h (d, ()), (* extend the data array *)
      __successor [f] [data] addr         (* increment its end address *)
    )

(* Creating an index to the end of an array. This index becomes valid as soon
   as the array is extended with at least one data cell. *)

term array_end_index : forall f. array f -> forall data. index (f `snoc` data) data =
  /\ f1.
  \a : array f1.
  let (_, addr) = a in
  /\ data.
    (* turning the current end address of the array into an index is just a matter
       of using [__array_read] and [__array_write]. *)
    (
      /\ f2. \a : array (f1 `snoc` data `@` f2). __array_read  [f1] [data] [f2] a addr,
      /\ f2. \a : array (f1 `snoc` data `@` f2). __array_write [f1] [data] [f2] a addr
    )

(* An index that is valid in a smaller array is also valid in a larger array. *)

term index_monotonic : forall f1 f2 data. index f1 data -> index (f1 `@` f2) data =
  /\ f1 f2 data.
  \i : index f1 data.
    let (get, set) = i in
    (
      /\ f3. get [f2 `@` f3],
      /\ f3. set [f2 `@` f3]
    )

(* Reading an array at an index. *)

term array_read : forall f data. array f -> index f data -> data =
  /\ f data.
  \a : array f.
  \i : index f data.
    let (get, _) = i in
    get [fnil] a
    (* trivial, since an index is a pair of [read] and [write] functions *)

(* Writing an array at an index. *)

term array_write : forall f data. array f -> index f data -> data -> array f =
  /\ f data.
  \a : array f.
  \i : index f data.
    let (_, set) = i in
    set [fnil] a
    (* trivial, since an index is a pair of [read] and [write] functions *)

(* ------------------------------------------------------------------------- *)

(* Private type definitions. *)

(* A heap (the array data) is just a tuple. A fragment is turned into a heap
   by plugging the unit type in its hole. *)

type __fheap : fragment -> * =
    \f. f ()

(* The end address of a fragment is represented as a pair of [get] and
   [set] functions. These functions allow reading and updating the data
   in the hole. *)

type __fget : fragment -> * =
  \f. forall tail. f tail -> tail

type __fset : fragment -> * =
  \f. forall tail1 tail2. f tail1 -> tail2 -> f tail2

type __faddress : fragment -> * =
  \f. (__fget f, __fset f)

(* ------------------------------------------------------------------------- *)

(* Private term definitions. *)

(* The address zero. *)

term __zero_get : __fget fnil =
  /\ tail. \x : tail. x

term __zero_set : __fset fnil =
  /\ tail1 tail2. \x : tail1. \y : tail2. y

term __zero : __faddress fnil =
  (
    __zero_get,
    __zero_set
  )

(* The successor of an address. Since an address is encoded as a pair of
   [get] and [set] functions, taking its successor involves building a new
   pair of such functions. *)

term __successor_get : forall f data. __fget f -> __fget (f `snoc` data) =
  /\ f data.
  \get : __fget f.
  /\ tail.
  \h : f (data, tail).
  let (_, tail) = get [fcell data tail] h in
  tail

term __successor_set : forall f data. __fget f -> __fset f -> __fset (f `snoc` data) =
  /\ f data.
  \get : __fget f.
  \set : __fset f.
  /\ tail1 tail2.
  \h : f (data, tail1).
  \x : tail2.
  let (data, _) = get [fcell data tail1] h in
  set [fcell data tail1] [fcell data tail2] h (data, x)

term __successor : forall f data. __faddress f -> __faddress (f `snoc` data) =
  /\ f data.
  \addr : __faddress f.
  let (get, set) = addr in
  (__successor_get [f] [data] get, __successor_set [f] [data] get set)

(* The functions [__array_read] and [__array_write] allow reading and writing
   an array at a certain address. The type of the array must be split up in
   precisely the right manner -- a fragment [f1], followed by a cell [data],
   followed by a fragment [f2] -- and the address must correspond to the end
   of the fragment [f1]. Then, the value that is read or written has type
   [data]. *)

term __array_read: forall f1 data f2. array (f1 `snoc` data `@` f2) -> __faddress f1 -> data =
  /\ f1 data f2.
  type df2 = fcell data `@` f2 in
  type hdf2 = __fheap df2 in
  \a : array (f1 `@` df2).
  \addr : __faddress f1.
    let (h, _) = a in
    let (get, _) = addr in
    let (d, _) = get [hdf2] h in
    d

term __array_write: forall f1 data f2. array (f1 `snoc` data `@` f2) -> __faddress f1 -> data ->
                                       array (f1 `snoc` data `@` f2) =
  /\ f1 data f2.
  type df2 = fcell data `@` f2 in
  type hdf2 = __fheap df2 in
  \a : array (f1 `@` df2).
  \addr : __faddress f1.
  \d : data.  
    let (h, end_addr) = a in
    let (get, set) = addr in
    let (_, tail) = get [hdf2] h in
    (
      set [hdf2] [hdf2] h (d, tail),
      end_addr
    )

