(**************************************************************************)
(*                                                                        *)
(*  FORK                                                                  *)
(*                                                                        *)
(*  Author:  Franois Pottier, INRIA Paris-Rocquencourt                   *)
(*  Version: 20101109                                                     *)
(*                                                                        *)
(*  The copyright to this code is held by Institut National de Recherche  *)
(*  en Informatique et en Automatique (INRIA). All rights reserved. This  *)
(*  file is distributed under the license CeCILL-C (see file LICENSE).    *)
(*                                                                        *)
(**************************************************************************)

(* The encoding of types. *)

(* A type in the monadic source language is translated to a semantic type.
   We do not define the translation as a monolithic function -- that would
   require induction over the structure of source types -- but we provide
   the combinators that represent each of the cases. *)

(* ------------------------------------------------------------------------- *)

(* A semantic type is a world-parameterized type. *)

kind stype = later world -> *

(* ------------------------------------------------------------------------- *)

(* Conventions. *)

default kind a : stype

default kind b : stype

default kind body : stype -> stype

(* ------------------------------------------------------------------------- *)

(* The necessity modality. *)

(* A value has semantic type [box t] in the current world if it has semantic
   type [t] in every future world. *)

type box : stype -> stype =
  \a x.
    forall y. a (x `o` y)

(* ------------------------------------------------------------------------- *)

(* Unit. *)

type unit : stype =
  \x. ()

(* ------------------------------------------------------------------------- *)

(* Pairs. *)

type pair : stype -> stype -> stype =
  \a b x. (a x, b x)

(* ------------------------------------------------------------------------- *)

(* Universal quantification. *)

type univ : (stype -> stype) -> stype =
  \body x.
    forall a. body a x

(* ------------------------------------------------------------------------- *)

(* Functions. *)

(* The argument of a function must be necessary, and its result is guaranteed
   to be necessary. *)

type arrow : stype -> stype -> stype =
  \a b x. box a x -> box b x

(* ------------------------------------------------------------------------- *)

(* The monad. *)

(* A computation is a function that maps a store in world [x] to a pair of
   a (necessary) value and a store in some future world [x `o` y]. *)

(* In the definition of [monad], it is important to note that the types [store x]
   and [store (x `o` y)] have kind [later *], not [*]. In order to prove that the
   arrow [store x -> outcome a x] has kind [*], as we are required, we exploit the
   fact that the type operators [->] and [(,)] are contractive, that is, they have
   kind [later * -> later * -> *]. *)

type monad : stype -> stype =
  \a x.
    store x -> outcome a x

type outcome : stype -> stype =
  \a x.
    exists y. (box a (x `o` y), store (x `o` y))

(* ------------------------------------------------------------------------- *)

(* Cells. *)

(* A cell is a world of width one. More precisely, if [a] is a semantic type
   and [x] is a world, then [cell a x] is a world, which describes a single
   memory cell, holding a value of type [a]. The parameter [x] represents the
   world before the cell is allocated. (We intend to initialize the cell with
   a value of type [box a x].) Thus, the world after the cell is allocated is
   [x `o` cell a x], and the world after an arbitrary world extension [y] has
   taken place is [x `o` cell a x `o` y]. This world is used as an argument to
   [a] in the definition of [cell]. Of course, this makes the definition
   recursive. *)

(* One could give a definition of [cell] that does not involve the parameter
   [x], and use [cell (a `tensor` x)] to recover the effect of the definition
   below. This was my original approach. It seems more difficult to explain. *)

type cell : stype -> later world -> world =
  \a x y tail.
    (a (x `o` cell a x `o` y), tail)

(* The following lemma helps understand what happens when composing a
   world and a cell. *)

lemma compose_cell:
  forall a w x.
  (w `o` cell a w) x =
  w (cell a w `o` x) `snoc` a (w `o` cell a w `o` x)

(* ------------------------------------------------------------------------- *)

(* Coercions. *)

(* The following functions have identity semantics, i.e., they have no runtime
   effect. This could be formally checked if the prototype implementation
   viewed coercions as a separate syntactic category. For now, it does not. *)

(* ------------------------------------------------------------------------- *)

(* A necessary value that is valid now remains valid later. *)

term forward :
  forall a x y.
  box a x ->
  box a (x `o` y)
=
  /\ a x y.
  \v : box a x.
  /\ z.
  v [ y `o` z ]

term box_boxbox :
  forall a x.
  box a x ->
  box (box a) x
=
  /\ a x.
  \v : box a x.
  /\ y z.
    v [ y `o` z ]

(* ------------------------------------------------------------------------- *)

(* An outcome that is valid later is valid now. *)

term backward :
  forall a x y.
  outcome a (x `o` y) ->
  outcome a x
=
  /\ a x y.
  \outcome : outcome a (x `o` y).
  unpack z, vs = outcome in
  pack y `o` z, vs as outcome a x

(* ------------------------------------------------------------------------- *)

(* The unit value is valid everywhere and forever. *)

term box_unit:
  forall x. box unit x
=
  /\ x y. ()

(* ------------------------------------------------------------------------- *)

(* A necessary pair is a pair of necessary values, and vice-versa. *)

term box_pair:
  forall a b x.
  box (pair a b) x ->
  (box a x, box b x)
=
  /\ a b x.
  \v : box (pair a b) x.
  (
    /\ x. let (v1, _) = v [ x ] in v1,
    /\ x. let (_, v2) = v [ x ] in v2
  )

term pair_box:
  forall a b x.
  (box a x, box b x) ->
  box (pair a b) x
=
  /\ a b x.
  \v : (box a x, box b x).
  let (v1, v2) = v in
  /\ y. (v1 [y], v2 [y])

(* ------------------------------------------------------------------------- *)

(* A necessary polymorphic value is a polymorphic necessary value, and vice-versa. *)

term box_univ:
  forall body x.
  box (univ body) x ->
  forall a. box (body a) x
=
  /\ body x.
  \v : box (univ body) x.
  /\ a y.
  v [ y ] [ a ]

term univ_box:
  forall body x.
  (forall a. box (body a) x) ->
  box (univ body) x
=
  /\ body x.
  \v : forall a. box (body a) x.
  /\ y a.
    v [ a ] [ y ]

